
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <usb.h>
#include <time.h>

/* vendor and product id */
#define USB_VID  0x0403
#define USB_PID  0xc630

/* command format:
 * 7 6 5 4 3 2 1 0
 * D D D C C C L L
 *
 * DDD = Device
 * CCC = Command
 * LL  = number of bytes in transfer - 1
 */


/* Devices */
#define BLIT_BOARD		(0<<5)
#define LCD_BOARD		(1<<5)
#define LM75_BOARD		(2<<5)
#define RTC_BOARD		(3<<5)

/* BLIT_BOARD commands */
#define BLIT_BOARD_ECHO		(BLIT_BOARD | (0<<2))
#define BLIT_BOARD_SET_LED	(BLIT_BOARD | (1<<2))

/* LM75_BOARD commands */
#define LM75_BOARD_STATUS	(LM75_BOARD | (0<<2))
#define LM75_BOARD_READ		(LM75_BOARD | (1<<2))

/* RTC_BOARD commands */
#define RTC_BOARD_STATUS	(RTC_BOARD | (0<<2))
#define RTC_BOARD_READ		(RTC_BOARD | (1<<2))
#define RTC_BOARD_WRITE		(RTC_BOARD | (2<<2))

/* LCD_Board commands */
#define LCD_BOARD_OUT_TEXT	(LCD_BOARD | (0<<2))
#define LCD_BOARD_INTERN	(LCD_BOARD | (1<<2))
#define LCD_BOARD_CLEAR		(LCD_BOARD | (2<<2))
#define LCD_BOARD_BACKLIGHT	(LCD_BOARD | (3<<2))

/* commandline */
#define CMD_ECHO		"echo"
#define CMD_SET_LED		"led"
#define CMD_LM75_READ		"lm75_read"
#define CMD_LM75_STATUS		"lm75_status"
#define CMD_RTC_READ		"rtc_read"
#define CMD_RTC_WRITE		"rtc_write"
#define CMD_RTC_GET_TIME	"rtc_get_time"
#define CMD_RTC_SET_TIME	"rtc_set_time"
#define CMD_RTC_STATUS		"rtc_status"
#define CMD_LCD_INTERN		"lcd_intern"
#define CMD_LCD_CLEAR		"lcd_clear"
#define CMD_LCD_BACKLIGHT	"lcd_backlight"
#define CMD_LCD_OUT_TEXT	"lcd_out_text"


#define FROMBCD(x)      (((x) >> 4) * 10 + ((x) & 0xf))
#define TOBCD(x)        (((x) / 10 * 16) + ((x) % 10))

usb_dev_handle      *handle = NULL;


//****************************************
//* Hilfeausgabe
void prog_help(void)
{
	printf("Syntax:\n");
	printf("=======\n");
	printf("usb2blitboard echo\n");
	printf("usb2blitboard led <0|1>\n");
	printf("usb2blitboard lm75_read\n");
	printf("usb2blitboard lm75_status\n");
	printf("usb2blitboard rtc_read <adresse>\n");
	printf("usb2blitboard rtc_write <adresse> <value>\n");
	printf("usb2blitboard rtc_status\n");
	printf("usb2blitboard rtc_get_time\n");
	printf("usb2blitboard rtc_set_time\n");
	printf("usb2blitboard lcd_intern <0|1>\n");
	printf("usb2blitboard lcd_clear\n");
	printf("usb2blitboard lcd_backlight <0|1>\n");
	printf("usb2blitboard lcd_out_text <y> <x> <text>\n");
}


/* ---------------------------------------------------------------------- */
int myusb_send(int request, int value, int index) {
  if(usb_control_msg(handle, USB_TYPE_VENDOR, request, 
		      value, index, NULL, 0, 1000) < 0) {
    fprintf(stderr, "send->USB request failed!\n");
    return -1;
  }
  return 0;
}

/* to increase performance, a little buffer is being used to */
/* collect command bytes of the same type before transmitting them */
#define BUFFER_MAX_CMD 4        /* current protocol supports up to 4 bytes */
int buffer_current_type = -1;   /* nothing in buffer yet */
int buffer_current_fill = 0;    /* -"- */
unsigned char buffer[BUFFER_MAX_CMD];


/* ---------------------------------------------------------------------- */
/* flush command queue due to buffer overflow / content */
/* change or due to explicit request */
void myusb_flush(void) {
  int request, value, index;
  
  /* anything to flush? ignore request if not */
  if (buffer_current_type == -1)
    return;
  
  /* build request byte */
  request = buffer_current_type | (buffer_current_fill - 1);
  
  /* fill value and index with buffer contents. endianess should IMHO not */
  /* be a problem, since usb_control_msg() will handle this. */
  value = buffer[0] | (buffer[1] << 8);
  index = buffer[2] | (buffer[3] << 8);
  
  /* send current buffer contents */
  myusb_send(request, value, index);
  
  /* buffer is now free again */
  buffer_current_type = -1;
  buffer_current_fill = 0;
}

/* ---------------------------------------------------------------------- */
/* enqueue a command into the buffer */
void myusb_enqueue(int command_type, int value) {
  if ((buffer_current_type >= 0) && (buffer_current_type != command_type))
    myusb_flush();
  
  /* add new item to buffer */
  buffer_current_type = command_type;
  buffer[buffer_current_fill++] = value;
  
  /* flush buffer if it's full */
  if (buffer_current_fill == BUFFER_MAX_CMD)
    myusb_flush();
}

/* ---------------------------------------------------------------------- */
#define ECHO_NUM 100
void blit_board_echo(void) {

  int i, nBytes, errors=0;
  unsigned short val, ret;

  for(i=0;i<ECHO_NUM;i++) {
    val = rand() & 0xffff;
    
    nBytes = usb_control_msg(handle, 
	   USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, 
	   BLIT_BOARD_ECHO, val, 0, (char*)&ret, sizeof(ret), 1000);

    if(nBytes < 0) {
      fprintf(stderr, "echo->USB request failed (%s)!\n", (char*)&ret);
      return;
    }

    if(val != ret)
      errors++;
  }

  if(errors) 
    fprintf(stderr, "ERROR: %d out of %d echo transfers failed!\n", 
	    errors, ECHO_NUM);
  else 
    printf("Echo test successful!\n");
}

/* ---------------------------------------------------------------------- */
void blit_board_set_led(unsigned char on_off) {

  int nBytes;
  unsigned short ret;

    nBytes = usb_control_msg(handle, 
	   USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, 
	   BLIT_BOARD_SET_LED, on_off, 0, (char*)&ret, sizeof(ret), 1000);

    if(nBytes < 0) {
      fprintf(stderr, "USB request failed (%s)!\n", (char*)&ret);
      return;
    }
}

/* ---------------------------------------------------------------------- */
/* get a value from the interface */
int myusb_get(unsigned char cmd) {
  unsigned char       buffer[2];
  int                 nBytes;

  /* send control request and accept return value */
  nBytes = usb_control_msg(handle, 
	   USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, 
	   cmd, 0, 0, (char *)buffer, sizeof(buffer), 1000);

  if(nBytes < 0) {
    fprintf(stderr, "get->USB request failed!\n");
    return -1;
  }

  return buffer[0] + 256*buffer[1];
}

/* ---------------------------------------------------------------------- */
void lm75_board_read(void) {

	int ver = myusb_get(LM75_BOARD_READ);
	unsigned char h_byte;
	unsigned char l_byte;
	char * s;

	h_byte = ver&0xff;
	l_byte = ver>>8;

	if (h_byte >= 128) {			// wenn Bit7 gesetzt, dann negative Temp.
		h_byte = ~(h_byte);		// dann (lt. Datenblatt) Byte invertieren
		h_byte++;			// und plus 1
		s= "-";
	} else {
		s= "+";
	}
	if (l_byte >= 128) {			// wenn Bit7 gesetzt, dann ,5 sonst ,0
		l_byte=5;
	} else {
		l_byte=0;
	}

	printf("%s%d.%d°C\n", s, h_byte, l_byte);
}

/* ---------------------------------------------------------------------- */
void lm75_board_status(void) {
	int status = myusb_get(LM75_BOARD_STATUS);

	printf("Status LM75-Board: %d\n", status);
}

/* ---------------------------------------------------------------------- */
void rtc_board_status(void) {
	int status = myusb_get(RTC_BOARD_STATUS);

	printf("Status RTC-Board: %d\n", status);
}

/* ---------------------------------------------------------------------- */
unsigned short rtc_board_read(unsigned char adr) {
  int nBytes;
  unsigned short ret;

    nBytes = usb_control_msg(handle, 
	   USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, 
	   RTC_BOARD_READ, adr, 0, (char*)&ret, sizeof(ret), 1000);

    if(nBytes < 0) {
      fprintf(stderr, "USB request failed (%s)!\n", (char*)&ret);
      return 0;
    }
	return ret;
}

/* ---------------------------------------------------------------------- */
void rtc_board_get_time(void) {
	unsigned char ss, mm, hh, dd, mo;

	ss = FROMBCD(rtc_board_read(2));	// Sekunden
	mm = FROMBCD(rtc_board_read(3));	// Minuten
	hh = FROMBCD(rtc_board_read(4));	// Stunden
	dd = FROMBCD(rtc_board_read(5));	// Tag
	mo = FROMBCD(rtc_board_read(6));	// Monat
	printf("RTC-Board-Time: %02d.%02d. %02d:%02d:%02d\n", dd, mo, hh, mm, ss);

}

/* ---------------------------------------------------------------------- */
unsigned char rtc_board_write(unsigned char adr, unsigned char val) {
	int nBytes;
	int send = val * 256 + adr;
	unsigned short ret;

    nBytes = usb_control_msg(handle, 
	   USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, 
	   RTC_BOARD_WRITE, send, 0, (char*)&ret, sizeof(ret), 1000);

    if(nBytes < 0) {
      fprintf(stderr, "USB request failed (%s)!\n", (char*)&ret);
      return 1;
    }
	return ret;
}

/* ---------------------------------------------------------------------- */
unsigned char rtc_board_set_time(void) {
	struct tm *nt;
	time_t long_time;

	time(&long_time);
	nt = localtime(&long_time);

	if (rtc_board_write(2, TOBCD(nt->tm_sec))) return 1;		// Sekunden
	if (rtc_board_write(3, TOBCD(nt->tm_min))) return 1;		// Minuten
	if (rtc_board_write(4, TOBCD(nt->tm_hour))) return 1;		// Stunden
	if (rtc_board_write(5, TOBCD(nt->tm_mday))) return 1;		// Tag
	if (rtc_board_write(6, TOBCD(nt->tm_mon + 1))) return 1;	// Monat (0...11)

	return 0;
}

/* ---------------------------------------------------------------------- */
unsigned char lcd_board_clear(void) {
	int status = myusb_get(LCD_BOARD_CLEAR);

	return status;
}

/* ---------------------------------------------------------------------- */
void lcd_board_intern(unsigned char on_off) {
  int nBytes;
  unsigned short ret;

    nBytes = usb_control_msg(handle, 
	   USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, 
	   LCD_BOARD_INTERN, on_off, 0, (char*)&ret, sizeof(ret), 1000);

    if(nBytes < 0) {
      fprintf(stderr, "USB request failed (%s)!\n", (char*)&ret);
      return;
    }
}

/* ---------------------------------------------------------------------- */
void lcd_board_backlight(unsigned char on_off) {
  int nBytes;
  unsigned short ret;

    nBytes = usb_control_msg(handle, 
	   USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, 
	   LCD_BOARD_BACKLIGHT, on_off, 0, (char*)&ret, sizeof(ret), 1000);

    if(nBytes < 0) {
      fprintf(stderr, "USB request failed (%s)!\n", (char*)&ret);
      return;
    }
}

/* ---------------------------------------------------------------------- */
void lcd_board_out_char(unsigned char y, unsigned char x,  unsigned char *txt) {
	int nBytes;
	int send, i;
	unsigned short ret;

	for (i=0; i < strlen(txt); i++) {
		send = ((y*32 + (x+i)) * 256) + txt[i];
		nBytes = usb_control_msg(handle, 
			USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, 
			LCD_BOARD_OUT_TEXT, send, 0, (char*)&ret, sizeof(ret), 1000);

		if(nBytes < 0) {
			fprintf(stderr, "USB request failed (%s)!\n", (char*)&ret);
			return;
		}
	}	
	return;
}

/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
int main(int argc, char *argv[]) {
	struct usb_bus      *bus;
	struct usb_device   *dev;
	int i, x, y;

	printf("usb2blitboard; Uwe Berger, 2008\n");

	usb_init();
  	usb_find_busses();
	usb_find_devices();
  
	for(bus = usb_get_busses(); bus; bus = bus->next) {
		for(dev = bus->devices; dev; dev = dev->next) {
			if ((dev->descriptor.idVendor == USB_VID) && 
			    (dev->descriptor.idProduct == USB_PID)) {
				printf("Found BLIT-Board device on bus %s device %s.\n", 
					bus->dirname, dev->filename);
				/* open device */
				if(!(handle = usb_open(dev))) 
					fprintf(stderr, "Error: Cannot open BLIT-Board device: %s\n", 
						usb_strerror());
			break;
			}
		}
	}
  
	if(!handle) {
		fprintf(stderr, "Error: Could not find BLIT-Board device\n");
		exit(-1);
	}

	//* Befehlsverzweigung (1.Argument ist das Schluesselwort)
	if (argc > 1) 
	{
		if (strcmp(argv[1], CMD_LM75_READ) == 0) {
			lm75_board_read();
		}
		else if (strcmp(argv[1], CMD_LM75_STATUS) == 0) {
			lm75_board_status();
		}
		else if (strcmp(argv[1], CMD_ECHO) == 0) {
			blit_board_echo();
		}
		else if (strcmp(argv[1], CMD_RTC_STATUS) == 0) {
			rtc_board_status();
		}
		else if (strcmp(argv[1], CMD_RTC_READ) == 0) {
			if (argc > 2) {
				printf("Adress %d: %d\n", atoi(argv[2]), rtc_board_read(atoi(argv[2])));
			} else {
				printf("usb2blitboard %s -> zu wenig Argumente!\n", CMD_RTC_READ);
				prog_help();
			}
		}
		else if (strcmp(argv[1], CMD_RTC_WRITE) == 0) {
			if (argc > 3) {
				printf("RTC-Board-Write: %d\n", rtc_board_write(atoi(argv[2]), atoi(argv[3])));
			} else {
				printf("usb2blitboard %s -> zu wenig Argumente!\n", CMD_RTC_WRITE);
				prog_help();
			}
		}
		else if (strcmp(argv[1], CMD_RTC_GET_TIME) == 0) {
			rtc_board_get_time();
		}
		else if (strcmp(argv[1], CMD_RTC_SET_TIME) == 0) {
			printf ("%s Return-Code: %d\n", argv[1], rtc_board_set_time());
		}
		else if (strcmp(argv[1], CMD_LCD_INTERN) == 0) {
			if (argc > 2) {
				lcd_board_intern(atoi(argv[2]));
			} else {
				printf("usb2blitboard %s -> zu wenig Argumente!\n", CMD_LCD_INTERN);
				prog_help();
			}
		}
		else if (strcmp(argv[1], CMD_LCD_CLEAR) == 0) {
			printf ("%s Return-Code: %d\n", argv[1], lcd_board_clear());
		}
		else if (strcmp(argv[1], CMD_LCD_BACKLIGHT) == 0) {
			if (argc > 2) {
				lcd_board_backlight(atoi(argv[2]));
			} else {
				printf("usb2blitboard %s -> zu wenig Argumente!\n", CMD_LCD_BACKLIGHT);
				prog_help();
			}
		}
		else if (strcmp(argv[1], CMD_LCD_OUT_TEXT) == 0) {
			if (argc > 4) {
				// ab argv[4] steht der Text. Jedes Wort waere dann ein
				// eigenes Kommandozeilenargument...
				x = atoi(argv[3]);
				y = atoi(argv[2]);
				for (i=4; i<argc; i++) {
					lcd_board_out_char(y, x, (unsigned char *)argv[i]);
					if ((i+1)<argc) {
						x = x + strlen(argv[i]);
						lcd_board_out_char(y, x, (unsigned char *)" ");
						x++;
					}
				}
			} else {
				printf("usb2blitboard %s -> zu wenig Argumente!\n", CMD_LCD_BACKLIGHT);
				prog_help();
			}
		}
		else if (strcmp(argv[1], CMD_SET_LED) == 0) {
			if (argc > 2) {
				blit_board_set_led(atoi(argv[2]));
			} else {
				printf("usb2blitboard %s -> zu wenig Argumente!\n", CMD_SET_LED);
				prog_help();
			}
		} else {
			printf("usb2blitboard: unbekannter Befehl!\n");
			prog_help();
		}
	} else {
		printf("usb2blitboard: zu wenig Argumente!\n");
		prog_help();
	}

  usb_close(handle);

  return 0;
}
